import math
import random
import pybirdsreynolds.const as const
import pybirdsreynolds.params as params
import pybirdsreynolds.variables as variables
from scipy.spatial import cKDTree

birds = []
velocities = []


def limit_speed(vx, vy):
    speed = math.sqrt(vx * vx + vy * vy)
    if speed > params.MAX_SPEED:
        vx = (vx / speed) * params.MAX_SPEED
        vy = (vy / speed) * params.MAX_SPEED
    return vx, vy


def move_birds(with_move, translate):
    if not birds or birds == []:
        for _ in range(params.NUM_BIRDS):
            px = random.randint(
                const.MARGIN + variables.WIDTH_CONTROLS,
                params.WIDTH - const.MARGIN + variables.WIDTH_CONTROLS,
            )
            py = random.randint(const.MARGIN, params.HEIGHT - const.MARGIN)
            birds.append((px, py))
            angle = random.uniform(0, 2 * math.pi)
            speed = random.uniform(0, params.MAX_SPEED)
            vx = speed * math.cos(angle)
            vy = speed * math.sin(angle)
            velocities.append((vx, vy))
    else:
        if translate:
            for i in range(len(birds)):
                x, y = birds[i]
                if variables.HIDDEN:
                    birds[i] = (x - const.WIDTH_CONTROLS_DEFAULT, y)
                else:
                    birds[i] = (x + const.WIDTH_CONTROLS_DEFAULT, y)
        # Keep birds only if inside
        inside_points = []
        inside_velocities = []
        for (x, y), (vx, vy) in zip(birds, velocities):
            if (
                variables.WIDTH_CONTROLS + const.MARGIN
                <= x
                <= variables.WIDTH_CONTROLS + params.WIDTH - const.MARGIN
                and 0 + const.MARGIN <= y <= params.HEIGHT - const.MARGIN
            ):
                inside_points.append((x, y))
                inside_velocities.append((vx, vy))
        birds[:] = inside_points
        velocities[:] = inside_velocities
        variables.COUNT = len(birds)

        # Add birds if not enough
        if params.NUM_BIRDS > variables.COUNT:
            for _ in range(params.NUM_BIRDS - variables.COUNT):
                px = random.randint(
                    const.MARGIN + variables.WIDTH_CONTROLS,
                    params.WIDTH - const.MARGIN + variables.WIDTH_CONTROLS,
                )
                py = random.randint(const.MARGIN, params.HEIGHT - const.MARGIN)
                birds.append((px, py))

                angle = random.uniform(0, 2 * math.pi)
                speed = random.uniform(0, params.MAX_SPEED)
                vx = speed * math.cos(angle)
                vy = speed * math.sin(angle)
                velocities.append((vx, vy))

        # Delete birds if not enough
        elif params.NUM_BIRDS < variables.COUNT:
            for _ in range(variables.COUNT - params.NUM_BIRDS):
                idx = random.randint(0, len(birds) - 1)
                birds.pop(idx)
                velocities.pop(idx)

        if with_move:
            tree = cKDTree(birds)
            new_velocities = []
            for i, (x, y) in enumerate(birds):
                move_sep_x, move_sep_y = 0, 0
                move_align_x, move_align_y, move_align_x_tmp, move_align_y_tmp = (
                    0,
                    0,
                    0,
                    0,
                )
                move_coh_x, move_coh_y, move_coh_x_tmp, move_coh_y_tmp = 0, 0, 0, 0
                neighbors = 0
                vx, vy = velocities[i]
                if params.NEIGHBOR_RADIUS > 0 and not (
                    params.SEP_WEIGHT == 0
                    and params.ALIGN_WEIGHT == 0
                    and params.COH_WEIGHT == 0
                ):
                    neighbors_idx = tree.query_ball_point(
                        [x, y], params.NEIGHBOR_RADIUS
                    )
                    for j in neighbors_idx:
                        if i == j:
                            continue

                        x2, y2 = birds[j]
                        vx2, vy2 = velocities[j]
                        dx, dy = x - x2, y - y2
                        radius2 = params.NEIGHBOR_RADIUS**2
                        dist2 = dx * dx + dy * dy
                        if 0 < dist2 < radius2:
                            dist = math.sqrt(dist2)
                            # SEPARATION
                            # If a neighbor is too close, add a vector to move away from it (opposite direction of the neighbor).
                            move_sep_x += (x - x2) / dist
                            move_sep_y += (y - y2) / dist
                            # ALIGNMENT
                            # Add the neighbor's velocity so the bird tends to align with it.
                            # Division is done later
                            vx2, vy2 = velocities[j]
                            move_align_x_tmp += vx2
                            move_align_y_tmp += vy2
                            # COHESION
                            # Add the neighbor's position to later calculate an average point,
                            # so the bird moves toward the group's center.
                            # Division is done later
                            move_coh_x_tmp += x2
                            move_coh_y_tmp += y2
                            neighbors += 1

                    if neighbors > 0:
                        move_align_x = move_align_x_tmp / neighbors
                        move_align_y = move_align_y_tmp / neighbors
                        move_coh_x = move_coh_x_tmp / neighbors
                        move_coh_y = move_coh_y_tmp / neighbors
                        move_coh_x = move_coh_x - x
                        move_coh_y = move_coh_y - y

                    vx += (
                        params.SEP_WEIGHT * move_sep_x
                        + params.ALIGN_WEIGHT * move_align_x
                        + params.COH_WEIGHT * move_coh_x
                    )
                    vy += (
                        params.SEP_WEIGHT * move_sep_y
                        + params.ALIGN_WEIGHT * move_align_y
                        + params.COH_WEIGHT * move_coh_y
                    )

                # RANDOM
                speed = math.sqrt(vx**2 + vy**2)
                if params.RANDOM_SPEED != 0:
                    target_speed = params.MAX_SPEED / 2
                    sigma_percent = params.RANDOM_SPEED
                    adjust_strength = 0.05
                    sigma_base = (sigma_percent / 100) * params.MAX_SPEED
                    weight = (
                        4 * speed * (params.MAX_SPEED - speed) / (params.MAX_SPEED**2)
                    )
                    sigma = sigma_base * weight
                    delta_speed = random.gauss(0, sigma)
                    new_speed = speed + delta_speed
                    new_speed += (target_speed - new_speed) * adjust_strength
                    new_speed = max(0.1, min(params.MAX_SPEED, new_speed))
                    factor = new_speed / speed
                    vx *= factor
                    vy *= factor
                if params.RANDOM_ANGLE != 0:
                    angle = math.atan2(vy, vx)
                    angle += math.radians(
                        random.uniform(-1 * params.RANDOM_ANGLE, params.RANDOM_ANGLE)
                    )
                    speed = math.sqrt(vx**2 + vy**2)
                    vx = speed * math.cos(angle)
                    vy = speed * math.sin(angle)
                vx, vy = limit_speed(vx, vy)

                new_velocities.append((vx, vy))

            # Update positions
            new_points = []
            for (x, y), (vx, vy) in zip(birds, new_velocities):
                nx = x + vx
                ny = y + vy
                # Bounces
                while (
                    nx < const.MARGIN + variables.WIDTH_CONTROLS
                    or nx > params.WIDTH - const.MARGIN + variables.WIDTH_CONTROLS
                ):
                    if nx < const.MARGIN + variables.WIDTH_CONTROLS:
                        overshoot = (const.MARGIN + variables.WIDTH_CONTROLS) - nx
                        nx = (const.MARGIN + variables.WIDTH_CONTROLS) + overshoot
                        vx = abs(vx)
                    elif nx > params.WIDTH - const.MARGIN + variables.WIDTH_CONTROLS:
                        overshoot = nx - (
                            params.WIDTH - const.MARGIN + variables.WIDTH_CONTROLS
                        )
                        nx = (
                            params.WIDTH - const.MARGIN + variables.WIDTH_CONTROLS
                        ) - overshoot
                        vx = -abs(vx)
                while ny < const.MARGIN or ny > params.HEIGHT - const.MARGIN:
                    if ny < const.MARGIN:
                        overshoot = const.MARGIN - ny
                        ny = const.MARGIN + overshoot
                        vy = abs(vy)
                    elif ny > params.HEIGHT - const.MARGIN:
                        overshoot = ny - (params.HEIGHT - const.MARGIN)
                        ny = (params.HEIGHT - const.MARGIN) - overshoot
                        vy = -abs(vy)
                idx = birds.index((x, y))
                velocities[idx] = (vx, vy)
                new_points.append((nx, ny))
            birds[:] = new_points
    if birds:
        variables.AVG_X = sum(x for x, _ in birds) / len(birds)
        variables.AVG_Y = sum(y for _, y in birds) / len(birds)
        variables.AVG_SPEED_ABS = sum(
            math.sqrt(vx**2 + vy**2) for vx, vy in velocities
        ) / len(velocities)
        variables.AVG_VX = sum(vx for vx, _ in velocities) / len(velocities)
        variables.AVG_VY = sum(vy for _, vy in velocities) / len(velocities)



import types
import copy
import sys
import tkinter as tk
import pybirdsreynolds.const as const
import pybirdsreynolds.params as params
import pybirdsreynolds.variables as variables
import pybirdsreynolds.reynolds as reynolds
import pybirdsreynolds.draw as draw
from pybirdsreynolds.args import get_help_text
from pybirdsreynolds.draw import (
    draw_root,
    draw_status_overlays,
    draw_birds,
    draw_panels,
    draw_box,
    maximize_minimize,
    is_maximized,
    next_frame,
)
from pybirdsreynolds.reynolds import move_birds


def signal_handler(sig, frame):
    print("Interrupted! Closing application...")
    draw.root.destroy()
    sys.exit(0)


def on_click(l, sens):
    """
    Select a parameter from a clicked line and trigger navigation.

    - Finds the clicked parameter from `l`.
    - Builds the list of active params (`*_DOC` / `*_TEXT`).
    - Updates `params.SELECTED_INDEX`.
    - Calls `on_other_key` with `sens`.

    Args:
        l (str): Clicked line.
        sens (str): Key direction/action.
    """
    first_word = l.split()[0] if l.split() else None
    lines = [
        f"{name.removesuffix('_DOC'):15} :    {str(getattr(params, name.removesuffix('_DOC'))).split(maxsplit=1)[0]}"
        for name in vars(params)
        if name.endswith("_DOC") and getattr(params, f"{name[:-4]}_ACTIVATED") == 2
    ] + [
        f"{name.removesuffix('_TEXT'):15} :    {getattr(params, name)} [{getattr(params, name.replace('_TEXT', '_COMMAND'))}]"
        for name in vars(params)
        if name.endswith("_TEXT") and getattr(params, f"{name[:-5]}_ACTIVATED") == 2
    ]
    params.SELECTED_INDEX = next(
        (i for i, line in enumerate(lines) if line.split(":")[0].strip() == first_word),
        0,
    )
    on_other_key(types.SimpleNamespace(keysym=sens))


def _repeat(ligne, direction):
    on_click(ligne, direction)
    if variables.REPEATING["active"]:
        variables.REPEATING["job"] = draw.canvas.after(
            100, lambda: _repeat(ligne, direction)
        )


def start_repeat(ligne, direction):
    variables.REPEATING["active"] = True
    _repeat(ligne, direction)


def stop_repeat():
    variables.REPEATING["active"] = False
    if variables.REPEATING["job"]:
        draw.canvas.after_cancel(variables.REPEATING["job"])
        variables.REPEATING["job"] = None


def on_shift_press(event):
    variables.shift_pressed = True


def on_shift_release(event):
    variables.shift_pressed = False


def toggle_pause(event=None):
    variables.BLINK_STATE = True
    variables.PAUSED = not variables.PAUSED
    draw_panels(False, on_other_key, start_repeat, stop_repeat)
    draw_status_overlays()


def on_other_key(event):
    """
    Handle key events for parameter control and UI actions.

    - Shift modifies step size (×10).
    - Arrows: navigate or adjust selected param.
    - Special params: toggle TRIANGLES, COLOR, FREE, FONT_TYPE.
    - Triggers redraw/update (birds, UI panels, messages).
    - Supports commands: reset, regenerate, toggle FPS/pause/maximize,
      next frame, average display, documentation, hide panels.

    Args:
        event: Key event object with `keysym` and optional `state`.
    """
    # If SHIFT_ACTIVATED == 0, force Shift to be ignored
    if params.SHIFT_ACTIVATED == 0:
        shift_pressed = False
    else:
        # Detect if Shift is pressed
        shift_pressed = getattr(event, "state", None)
        if shift_pressed is not None:
            # bitmask: check Shift key
            shift_pressed = (shift_pressed & 0x1) != 0
        else:
            # fallback to stored state
            shift_pressed = variables.SHIFT_PRESSED

    # Multiplier: x10 when Shift is pressed, otherwise x1
    multiplier = 10 if shift_pressed else 1

    # Value: positive for "Right" key, negative for others (e.g. "Left")
    val = multiplier if event.keysym == "Right" else -multiplier

    # Get currently selected parameter from the UI order
    param = params.PARAM_ORDER_IHM[params.SELECTED_INDEX]

    # If the window is maximized, prevent changes for WIDTH or HEIGHT
    if param in ("WIDTH", "HEIGHT") and is_maximized():
        val = 0

    # Navigate through parameters using arrow keys if arrows are activated
    if params.ARROWS_ACTIVATED > 0 and not variables.HIDDEN:
        if event.keysym == "Up":
            params.SELECTED_INDEX = (params.SELECTED_INDEX - 1) % len(
                params.PARAM_ORDER_IHM
            )
        elif event.keysym == "Down":
            params.SELECTED_INDEX = (params.SELECTED_INDEX + 1) % len(
                params.PARAM_ORDER_IHM
            )

    # Change value with Right and Left
    if (
        event.keysym == "Right" or event.keysym == "Left"
    ) and params.ARROWS_ACTIVATED >= 1 and not variables.HIDDEN:
        # Specific cases
        if param == "TRIANGLES":
            variables.POINTS_ID = []
            draw.canvas.delete("bird")
            params.TRIANGLES = not params.TRIANGLES
        elif param == "FONT_TYPE":
            current_index = const.FONT_TYPE_LIST.index(params.FONT_TYPE)
            params.FONT_TYPE = const.FONT_TYPE_LIST[
                (current_index + val) % len(const.FONT_TYPE_LIST)
            ]
        elif param == "COLOR":
            variables.POINTS_ID = []
            draw.canvas.delete("bird")
            params.COLOR = not params.COLOR
            if not params.COLOR:
                variables.CANVAS_BG = "black"
                variables.FILL_COLOR = "white"
                variables.OUTLINE_COLOR = "white"
                variables.INFO_COLOR = "yellow"
            else:
                variables.CANVAS_BG = "#87CEEB"
                variables.FILL_COLOR = "black"
                variables.OUTLINE_COLOR = "black"
                variables.INFO_COLOR = "orange"
        elif param == "FREE":
            params.FREE = not params.FREE
            # Readjust params one by one if needed (no free)
            if not params.FREE:
                for paramm in params.PARAM_ORDER_IHM:
                    if paramm not in ["FREE", "COLOR", "TRIANGLES", "FONT_TYPE"]:
                        setattr(params, paramm, change_value(paramm, 0, params.FREE))
        else:
            setattr(params, param, change_value(param, val, params.FREE))

        # regenerate birds and ihm
        move_birds(False, False)
        draw_root()
        draw_panels(True, on_other_key, start_repeat, stop_repeat)
        draw_birds()
        draw_box()
        draw_status_overlays()
    # Reset case
    elif (
        getattr(event, "keysym", "").lower() == str(params.RESET_COMMAND)
        and params.RESET_ACTIVATED >= 1
    ):
        restore_options()
        move_birds(False, False)
        draw_panels(False, on_other_key, start_repeat, stop_repeat)
        draw_birds()
        draw_box()
        draw_status_overlays()
        draw_root()
        draw_panels(True, on_other_key, start_repeat, stop_repeat)
        draw.root.state("normal")
        draw.root.focus_force()
        draw.root.focus_set()
    # Rebirds
    elif (
        getattr(event, "keysym", "").lower() == str(params.REGENERATION_COMMAND)
        and params.REGENERATION_ACTIVATED >= 1
    ):
        reynolds.velocities = []
        reynolds.birds = []
        move_birds(False, False)
        draw_birds()
    # Toggle FPS
    elif (
        getattr(event, "keysym", "").lower() == str(params.TOGGLE_FPS_COMMAND)
        and params.TOGGLE_FPS_ACTIVATED >= 1
    ):
        variables.FPS = not variables.FPS
        draw_status_overlays()
    # Toggle Start Pause
    elif (
        getattr(event, "keysym", "").lower() == str(params.TOGGLE_START_PAUSE_COMMAND)
        and params.TOGGLE_START_PAUSE_ACTIVATED >= 1
    ):
        toggle_pause()
    # Next Frame
    elif (
        getattr(event, "keysym", "").lower() == str(params.NEXT_FRAME_COMMAND)
        and params.NEXT_FRAME_ACTIVATED >= 1
    ):
        next_frame()
    # Toggle Minimimize Maximize
    elif (
        getattr(event, "keysym", "").lower() == str(params.TOGGLE_MAXIMIZE_COMMAND)
        and params.TOGGLE_MAXIMIZE_ACTIVATED >= 1
    ):
        maximize_minimize(False)
    # Toggle Average Positon
    elif (
        getattr(event, "keysym", "").lower() == str(params.AVERAGE_COMMAND)
        and params.AVERAGE_ACTIVATED >= 1
    ):
        variables.AVERAGE = not variables.AVERAGE
        draw_status_overlays()
    # Killer mode
    elif (
        getattr(event, "keysym", "").lower() == str(params.HUNTER_COMMAND)
        and params.HUNTER_ACTIVATED >= 1
    ):
        variables.HUNTER = not variables.HUNTER
        draw_status_overlays()
        if variables.HUNTER == True:
            draw.root.bind("<Button-1>", on_left_click)
        else:
            draw.root.unbind("<Button-1>")
    # Documentation display
    elif (
        getattr(event, "keysym", "").lower() == str(params.DOC_COMMAND)
        and params.DOC_ACTIVATED >= 1
    ):
        help = f"pybirdsreynolds {const.VERSION_PROG}\n\n" + get_help_text()
        popin = tk.Toplevel(draw.canvas)
        popin.title("Documentation - pybirdreynolds 🐦")
        popin.transient(draw.canvas.winfo_toplevel())
        popin.geometry("+200+200")
        popin.configure(bg=variables.CANVAS_BG)

        frame = tk.Frame(popin, bg=variables.CANVAS_BG)
        frame.pack(padx=10, pady=10)

        text_widget = tk.Text(
            frame,
            wrap="word",
            width=80,
            height=20,
            bg=variables.CANVAS_BG,
            fg=variables.FILL_COLOR,
            insertbackground=variables.CANVAS_BG,
            highlightthickness=0,
            bd=0,
        )
        text_widget.insert("1.0", help)
        text_widget.config(state="disabled")
        text_widget.pack(side="left", fill="both", expand=True)

        scrollbar = tk.Scrollbar(frame, command=text_widget.yview)
        scrollbar.pack(side="right", fill="y")
        text_widget.config(yscrollcommand=scrollbar.set)
        popin.wait_visibility()
        popin.grab_set()
    # Toggle Hide Command
    elif (
        getattr(event, "keysym", "").lower() == str(params.HIDE_COMMAND)
        and params.HIDE_ACTIVATED >= 1
    ):
        if not variables.HIDDEN:
            variables.HIDDEN = True
            if is_maximized():
                maximize_minimize(True)
            variables.TRANS_HIDDEN = True
            variables.WIDTH_PARAMS = 0
            variables.WIDTH_CONTROLS = 0
            variables.HEIGHT_PARAMS_CONTROLS = 0
            params.HEIGHT_MIN=50
            params.HEIGHT_FREE_MIN=50
            params.WIDTH_MIN=50
            params.WIDTH_FREE_MIN=50
            draw_panels(True, on_other_key, start_repeat, stop_repeat)
            draw_root()
            draw_panels(False, on_other_key, start_repeat, stop_repeat)
            draw_box()
            draw_status_overlays()
            move_birds(False, True)
            draw_birds()
        else:
            variables.HIDDEN = False
            variables.WIDTH_PARAMS = const.WIDTH_PARAMS_DEFAULT
            variables.WIDTH_CONTROLS = const.WIDTH_CONTROLS_DEFAULT
            variables.HEIGHT_PARAMS_CONTROLS = const.HEIGHT_PARAMS_CONTROLS_DEFAULT
            if is_maximized():
                params.WIDTH = (
                    draw.root.winfo_width()
                    - variables.WIDTH_PARAMS
                    - variables.WIDTH_CONTROLS
                )
            params.HEIGHT_MIN=500
            params.HEIGHT_FREE_MIN=500
            params.WIDTH_MIN=500
            params.WIDTH_FREE_MIN=500    
            if params.HEIGHT<params.HEIGHT_MIN:
                params.HEIGHT=params.HEIGHT_MIN
            if params.WIDTH<params.WIDTH_MIN:
                params.WIDTH=params.WIDTH_MIN                
            draw_root()
            move_birds(False, True)
            draw_birds()
            draw_status_overlays()
            draw_panels(True, on_other_key, start_repeat, stop_repeat)
            draw_box()
            draw_root()
            draw_box()
    draw_panels(False, on_other_key, start_repeat, stop_repeat)

def on_left_click(event):
    change=False
    if not variables.HUNTER:
        return  # si le mode chasseur n'est pas actif, on ne fait rien

    # Distance maximum du clic pour "tuer" un oiseau
    kill_radius = params.HUNTER_SIZE

    # On crée une nouvelle liste d'oiseaux et d'IDs conservés
    new_birds = []
    new_velocities = []
    new_points_id = []

    for pid, (x, y), vel in zip(variables.POINTS_ID, reynolds.birds, reynolds.velocities):
        dist = ((event.x - x) ** 2 + (event.y - y) ** 2) ** 0.5

        if dist <= kill_radius:
            change=True
            draw.canvas.delete(pid)
            params.NUM_BIRDS=params.NUM_BIRDS-1
        else:
            new_birds.append((x, y))
            new_velocities.append(vel)
            new_points_id.append(pid)

    reynolds.birds = new_birds
    reynolds.velocities = new_velocities
    variables.POINTS_ID = new_points_id    
    if change:
        draw_panels(False, on_other_key, start_repeat, stop_repeat)

def restore_options():

    params.MAX_SPEED = copy.deepcopy(params.MAX_SPEED_INIT)
    params.NEIGHBOR_RADIUS = copy.deepcopy(params.NEIGHBOR_RADIUS_INIT)
    params.NUM_BIRDS = copy.deepcopy(params.NUM_BIRDS_INIT)
    params.WIDTH = copy.deepcopy(params.WIDTH_INIT)
    params.HEIGHT = copy.deepcopy(params.HEIGHT_INIT)
    params.REFRESH_MS = copy.deepcopy(params.REFRESH_MS_INIT)
    params.RANDOM_SPEED = copy.deepcopy(params.RANDOM_SPEED_INIT)
    params.RANDOM_ANGLE = copy.deepcopy(params.RANDOM_ANGLE_INIT)
    params.SEP_WEIGHT = copy.deepcopy(params.SEP_WEIGHT_INIT)
    params.ALIGN_WEIGHT = copy.deepcopy(params.ALIGN_WEIGHT_INIT)
    params.COH_WEIGHT = copy.deepcopy(params.COH_WEIGHT_INIT)
    params.SIZE = copy.deepcopy(params.SIZE_INIT)
    params.COLOR = copy.deepcopy(params.COLOR_INIT)
    params.FREE = copy.deepcopy(params.FREE_INIT)
    params.TRIANGLES = copy.deepcopy(params.TRIANGLES_INIT)
    params.FONT_TYPE = copy.deepcopy(params.FONT_TYPE_INIT)

    if not params.COLOR:
        variables.CANVAS_BG = "black"
        variables.FILL_COLOR = "white"
        variables.OUTLINE_COLOR = "white"
        variables.INFO_COLOR = "yellow"
    else:
        variables.CANVAS_BG = "#87CEEB"
        variables.FILL_COLOR = "black"
        variables.OUTLINE_COLOR = "black"
        variables.INFO_COLOR = "orange"
    variables.POINTS_ID = []
    draw.canvas.delete("bird")


def change_value(type, val, free):
    """
    Update a parameter value by adding `val` and clamp it within defined bounds.

    Bounds are taken from attributes in `params`:
      - Normal mode (`free=False`): <TYPE>_MIN / <TYPE>_MAX
      - Free mode   (`free=True`):  <TYPE>_FREE_MIN / <TYPE>_FREE_MAX

    Args:
        type (str): Parameter name (e.g., "speed").
        val (float): Increment to apply.
        free (bool): Whether to use free-mode bounds.

    Returns:
        float: The updated, bounded value.
    """
    value = getattr(params, type, None)
    prefix = type.upper()
    min_value = getattr(params, f"{prefix}_MIN", None)
    max_value = getattr(params, f"{prefix}_MAX", None)
    min_free_value = getattr(params, f"{prefix}_FREE_MIN", None)
    max_free_value = getattr(params, f"{prefix}_FREE_MAX", None)
    value += val
    if not free:
        if max_value is not None:
            value = min(value, max_value)
        if min_value is not None:
            value = max(value, min_value)
    else:
        if max_free_value is not None:
            value = min(value, max_free_value)
        if min_free_value is not None:
            value = max(value, min_free_value)
    return value



import time
import math
import types
import tkinter as tk
import pybirdsreynolds.const as const
import pybirdsreynolds.params as params
import pybirdsreynolds.variables as variables
import pybirdsreynolds.reynolds as reynolds
from tkinter import font
from pybirdsreynolds.reynolds import move_birds
from pybirdsreynolds.args import display_range

root = None
canvas = None


def patch_1():

    # TODO BUGFIX
    root.geometry(
        f"{variables.WIDTH_PARAMS + params.WIDTH +1+ variables.WIDTH_CONTROLS}x{max(params.HEIGHT -1, variables.HEIGHT_PARAMS_CONTROLS)}"
    )


def patch_2():
    # TODO BUGFIX
    root.geometry(
        f"{variables.WIDTH_PARAMS + params.WIDTH +3+ variables.WIDTH_CONTROLS}x{max(params.HEIGHT -3, variables.HEIGHT_PARAMS_CONTROLS)}"
    )


def next_frame():
    if variables.PAUSED:
        move_birds(True, False)
        draw_birds()


def update():
    """
    Main loop: update simulation, redraw, and track FPS.

    - If not paused: move birds, redraw, count frames, compute FPS.
    - If paused: reset FPS counters.
    - Reschedules itself with `root.after`.

    Returns:
        None
    """
    if not variables.PAUSED:
        move_birds(True, False)
        draw_birds()
        draw_status_overlays()
        variables.FRAME_COUNT += 1
        now = time.time()
        if not variables.COUNT:
            variables.LAST_TIME = now
            variables.COUNT = True
        # to stabilize fps
        if now - variables.LAST_TIME >= 1.0:
            variables.FPS_VALUE = variables.FRAME_COUNT / (now - variables.LAST_TIME)
            variables.FRAME_COUNT = 0
            variables.LAST_TIME = now
    # reset fps if paused
    else:
        variables.FRAME_COUNT = 0
        variables.COUNT = False
        variables.FPS_VALUE = 0

    root.after(params.REFRESH_MS, update)


def draw_panels(fullRefreshControls, on_other_key, start_repeat, stop_repeat):
    """
    Draw/update parameter and control panels on the canvas.

    - Build parameter/control lines from `params`.
    - Highlight selected param and attach tooltips.
    - If `fullRefreshControls=True`, create or update buttons (params & controls).
    - Handles button bindings for click/repeat navigation.
    - Clears everything if panels are hidden.
    """
    normal_font = font.Font(
        family=params.FONT_TYPE, size=params.FONT_SIZE, weight="normal"
    )

    # Build display lines:
    # - *_DOC → params with a docstring
    # - *_TEXT → control buttons with label + command
    lines = [
        f"{name.removesuffix('_DOC'):15} :     {str(getattr(params, name.removesuffix('_DOC'))).split(maxsplit=1)[0]}"
        for name in vars(params)
        if name.endswith("_DOC") and getattr(params, f"{name[:-4]}_ACTIVATED") == 2
    ] + [
        f"{name.removesuffix('_TEXT'):15} : {getattr(params, name)} [{getattr(params, name.replace('_TEXT', '_COMMAND'))}]"
        for name in vars(params)
        if name.endswith("_TEXT") and getattr(params, f"{name[:-5]}_ACTIVATED") == 2
    ]
    # Initial offsets
    x_text = 10
    y_text = 10
    # Clear previous drawings
    canvas.delete("controls")
    canvas.delete("params")

    # If panels are hidden → clear buttons and return
    if variables.WIDTH_PARAMS == 0 and variables.WIDTH_CONTROLS == 0:
        canvas.delete("params_button")
        canvas.delete("controls_button")
        for name in dir(params):
            if name.endswith(("_BUTTON", "_BUTTON_UP", "_BUTTON_DOWN")):
                setattr(params, name, None)
        return

    i_param = -1
    i_control = -1
    y_pos_control = 0

    # Iterate through each line to draw parameters & controls
    for i, line in enumerate(lines):
        key = line.split()[0]
        font_to_use = normal_font
        fill = variables.FILL_COLOR

        # Highlight selected param if arrows navigation is active
        if i == params.SELECTED_INDEX and params.ARROWS_ACTIVATED > 0:
            i_param = i_param + 1
            fill = "red"
            item = canvas.create_text(
                x_text + variables.WIDTH_CONTROLS + params.WIDTH,
                y_text + i_param * 2.3 * params.FONT_SIZE,
                anchor="nw",
                fill=fill,
                font=font_to_use,
                tags="params",
                text=line.lower(),
            )
            # Tooltip: docstring + value range
            add_canvas_tooltip(
                item,
                getattr(params, key.upper() + "_DOC")
                + " ("
                + display_range(key.upper())
                + ")",
            )
        # If it's a control line ([...]) -tooltip will be made later
        elif "[" in line:
            i_control = i_control + 1
        # Else → standard param
        elif not "[" in line:
            i_param = i_param + 1
            item = canvas.create_text(
                x_text + variables.WIDTH_CONTROLS + params.WIDTH,
                y_text + i_param * 2.3 * params.FONT_SIZE,
                anchor="nw",
                fill=variables.FILL_COLOR,
                font=font_to_use,
                tags="params",
                text=line.lower(),
            )
            add_canvas_tooltip(
                item,
                getattr(params, key.upper() + "_DOC")
                + " ("
                + display_range(key.upper())
                + ")",
            )

        # Track control & param vertical positions
        y_pos_control = y_text + i_control * 2.3 * 2 * params.FONT_SIZE
        y_pos_param = y_text + i_param * 2.3 * params.FONT_SIZE - 5

        # --- Full refresh: create/update buttons ---
        if fullRefreshControls:
            first_colon_index = line.find(":") + 1
            f = font.Font(font=font_to_use)
            x_offset = 0

            # Case: control button (with icon + command)
            if "[" in line:
                key = line.split()[0]
                btn_font = (params.FONT_TYPE, params.FONT_SIZE * 2)
                btn_width = 2
                btn_height = 1
                highlight_color = "black"
                highlight_thickness = 2

                name_button = key + "_BUTTON"
                key = line.split()[0]
                icon = getattr(params, key.upper() + "_ICON")
                cmd = getattr(params, key.upper() + "_COMMAND")

                lbl_btn_tmp = tk.Label(
                    canvas,
                    text=icon,
                    fg="black",
                    bg="white",
                    font=btn_font,
                    width=btn_width,
                    height=btn_height,
                    anchor="center",
                    highlightbackground=highlight_color,
                    highlightthickness=highlight_thickness,
                )

                tooltip_text = f"{getattr(params, key.upper() + '_TEXT')} [{getattr(params, key.upper() + '_COMMAND')}]"
                add_widget_tooltip(lbl_btn_tmp, tooltip_text)

                # Bind button → simulate key event
                lbl_btn_tmp.bind(
                    "<Button-1>",
                    lambda e, c=cmd: on_other_key(types.SimpleNamespace(keysym=c)),
                )

                # Create or reposition control button
                if getattr(params, name_button, None) is None:
                    setattr(
                        params,
                        name_button,
                        canvas.create_window(
                            x_text + x_offset + 2,
                            y_pos_control,
                            anchor="nw",
                            window=lbl_btn_tmp,
                            tags=("controls_button",),
                        ),
                    )
                else:
                    canvas.coords(
                        getattr(params, name_button),
                        x_text + x_offset + 2,
                        y_pos_control,
                    )

            # Case: param with +/- buttons (if arrows are active)
            first_colon_index = line.find(":") + 1
            f = font.Font(font=font_to_use)
            x_offset = f.measure(line[:first_colon_index])
            if "[" not in line and params.ARROWS_ACTIVATED > 0:
                highlight_color = "black"
                highlight_thickness = 1
                # LEFT/DOWN CASE
                name_button_down = key + "_BUTTON_DOWN"
                lbl_left = tk.Label(
                    canvas,
                    text="<",
                    fg="black",
                    bg="white",
                    font=font_to_use,
                    highlightbackground=highlight_color,
                    highlightthickness=highlight_thickness,
                )
                lbl_left.bind(
                    "<ButtonPress-1>", lambda e, l=line: start_repeat(l, "Left")
                )
                lbl_left.bind("<ButtonRelease-1>", lambda e: stop_repeat())
                # create op update button's position
                if (
                    not hasattr(params, name_button_down)
                    or getattr(params, name_button_down) is None
                ):
                    setattr(
                        params,
                        name_button_down,
                        canvas.create_window(
                            x_text
                            + x_offset
                            + 1
                            + variables.WIDTH_CONTROLS
                            + params.WIDTH,
                            y_pos_param,
                            anchor="nw",
                            window=lbl_left,
                            tags=("params_button",),
                        ),
                    )
                else:
                    canvas.coords(
                        getattr(params, name_button_down),
                        x_text + x_offset + 1 + variables.WIDTH_CONTROLS + params.WIDTH,
                        y_pos_param,
                    )

                # RIGHT/UP CASE
                name_button_up = key + "_BUTTON_UP"
                lbl_right = tk.Label(
                    canvas,
                    text=">",
                    fg="black",
                    bg="white",
                    font=font_to_use,
                    highlightbackground=highlight_color,
                    highlightthickness=highlight_thickness,
                )
                lbl_right.bind(
                    "<ButtonPress-1>", lambda e, l=line: start_repeat(l, "Right")
                )
                lbl_right.bind("<ButtonRelease-1>", lambda e: stop_repeat())
                # create op update button's position
                if (
                    not hasattr(params, name_button_up)
                    or getattr(params, name_button_up) is None
                ):
                    setattr(
                        params,
                        name_button_up,
                        canvas.create_window(
                            x_text
                            + x_offset
                            + 18
                            + variables.WIDTH_CONTROLS
                            + params.WIDTH,
                            y_pos_param,
                            anchor="nw",
                            window=lbl_right,
                            tags=("params_button",),
                        ),
                    )
                else:
                    canvas.coords(
                        getattr(params, name_button_up),
                        x_text
                        + x_offset
                        + 18
                        + variables.WIDTH_CONTROLS
                        + params.WIDTH,
                        y_pos_param,
                    )
    if variables.HUNTER == True:
        root.config(cursor="pirate")
        if hasattr(variables, "CURSOR_CIRCLE_ID"):
            canvas.delete(variables.CURSOR_CIRCLE_ID)
            del variables.CURSOR_CIRCLE_ID
        if hasattr(variables, "MOVE_CIRCLE_FUNC"):
            canvas.unbind("<Motion>")
            del variables.MOVE_CIRCLE_FUNC          
        r = params.HUNTER_SIZE
        cursor_circle = canvas.create_oval(
            0, 0, r * 2, r * 2,
            outline="red", width=2,
            tags="cursor_circle"
        )
        mouse_x = canvas.winfo_pointerx() - canvas.winfo_rootx()
        mouse_y = canvas.winfo_pointery() - canvas.winfo_rooty()
        x1, y1, x2, y2 = canvas.coords("boundary")
        if x1 <= mouse_x <= x2 and y1 <= mouse_y <= y2:
            canvas.coords(cursor_circle, mouse_x - r, mouse_y - r, mouse_x + r, mouse_y + r)
            canvas.tag_raise(cursor_circle)
            canvas.itemconfigure(cursor_circle, state="normal")
        else:
            canvas.itemconfigure(cursor_circle, state="hidden")
        def move_circle(event):
            x, y = event.x, event.y
            x1, y1, x2, y2 = canvas.coords("boundary")
            if x1 <= x <= x2 and y1 <= y <= y2:
                canvas.coords(cursor_circle, x - r, y - r, x + r, y + r)
                canvas.tag_raise(cursor_circle)
                canvas.itemconfigure(cursor_circle, state="normal")
            else:
                canvas.itemconfigure(cursor_circle, state="hidden")                

        variables.CURSOR_CIRCLE_ID = cursor_circle
        variables.MOVE_CIRCLE_FUNC = move_circle
        canvas.bind("<Motion>", move_circle)

    else:
        root.config(cursor="arrow")
        if hasattr(variables, "CURSOR_CIRCLE_ID"):
            canvas.delete(variables.CURSOR_CIRCLE_ID)
            del variables.CURSOR_CIRCLE_ID

        if hasattr(variables, "MOVE_CIRCLE_FUNC"):
            canvas.unbind("<Motion>")
            del variables.MOVE_CIRCLE_FUNC  

def draw_status_overlays():
    """
    Draw overlay messages on the canvas (paused, FPS, hidden panels, average point).

    - Shows "PAUSED" when simulation is paused.
    - Displays FPS value (or "NA/...").
    - Indicates when panels are hidden.
    - Marks average position if enabled.
    """
    canvas.delete("paused")
    if variables.PAUSED:
        canvas.create_text(
            variables.WIDTH_CONTROLS,
            max(params.HEIGHT, variables.HEIGHT_PARAMS_CONTROLS),
            anchor="sw",
            fill="red",
            font=(params.FONT_TYPE, params.FONT_SIZE, "bold"),
            tags="paused",
            text=" PAUSED ",
        )
    canvas.delete("fps")
    if variables.FPS:
        if not variables.PAUSED:
            if variables.FPS_VALUE == 0:
                value = "..."
            else:
                value = f"{variables.FPS_VALUE:.1f}"
        else:
            value = "NA"
        canvas.create_text(
            variables.WIDTH_CONTROLS,
            0,
            anchor="nw",
            fill="gray",
            font=(params.FONT_TYPE, params.FONT_SIZE, "bold"),
            tags="fps",
            text=f" FPS : {value}",
        )
    canvas.delete("hidden")
    if variables.HIDDEN:
        canvas.create_text(
            variables.WIDTH_CONTROLS + params.WIDTH,
            max(params.HEIGHT, variables.HEIGHT_PARAMS_CONTROLS),
            anchor="se",
            fill="gray",
            font=(params.FONT_TYPE, params.FONT_SIZE),
            tags="hidden",
            text="h to restore panels ",
        )
    canvas.delete("average")
    if variables.AVERAGE:
        canvas.create_oval(
            variables.AVG_X - 2 * params.SIZE,
            variables.AVG_Y - 2 * params.SIZE,
            variables.AVG_X + 2 * params.SIZE,
            variables.AVG_Y + 2 * params.SIZE,
            fill="red",
            outline="red",
            tag="average",
        )
    canvas.delete("hunter")
    if variables.HUNTER:
        canvas.create_text(
            variables.WIDTH_CONTROLS + params.WIDTH,
            0,
            anchor="ne",
            fill="red",
            font=(params.FONT_TYPE, params.FONT_SIZE),
            tags="hunter",
            text="Hunter mode ",
        )


def draw_box():
    if not is_maximized():
        const.WIDTH_BEFORE_MAXIMIZED = params.WIDTH
        const.HEIGHT_BEFORE_MAXIMIZED = params.HEIGHT
    canvas.delete("boundary")
    canvas.create_rectangle(
        variables.WIDTH_CONTROLS,
        0,
        variables.WIDTH_CONTROLS + params.WIDTH,
        params.HEIGHT,
        outline=variables.FILL_COLOR,
        width=const.MARGIN,
        tags="boundary",
    )

def draw_root():
    if not variables.HIDDEN:
        root.geometry(
            f"{variables.WIDTH_PARAMS+params.WIDTH+variables.WIDTH_CONTROLS+2}x{max(params.HEIGHT, variables.HEIGHT_PARAMS_CONTROLS)}"
        )
        canvas.config(
            width=variables.WIDTH_PARAMS + params.WIDTH + variables.WIDTH_CONTROLS + 2,
            height=max(params.HEIGHT, variables.HEIGHT_PARAMS_CONTROLS),
            bg=variables.CANVAS_BG,
        )
    else:
        root.geometry(
            f"{params.WIDTH}x{max(params.HEIGHT, variables.HEIGHT_PARAMS_CONTROLS)}"
        )
        root.minsize(params.WIDTH, params.HEIGHT)
        root.maxsize(params.WIDTH, params.HEIGHT)
        width_tmp = params.WIDTH
        height_tmp = params.HEIGHT
        root.update()
        root.minsize(params.WIDTH_MIN, params.HEIGHT_MIN)
        root.maxsize(10000, 10000)
        root.update()
        params.WIDTH = width_tmp
        params.HEIGHT = height_tmp

def is_maximized():
    if root.tk.call("tk", "windowingsystem") == "aqua":
        return bool(root.attributes("-fullscreen"))
    if root.state() == "zoomed":
        return True
    try:
        if root.attributes("-zoomed"):
            return True
    except tk.TclError:
        pass
    return (
        root.winfo_width() >= root.winfo_screenwidth()
        and root.winfo_height() >= root.winfo_screenheight()
    )


def draw_birds():
    fill = variables.FILL_COLOR
    outline = variables.OUTLINE_COLOR
    size = params.SIZE
    cos_150 = math.cos(math.radians(150))
    sin_150 = math.sin(math.radians(150))

    triangle_size = 6 * size
    triangle_width = 4 * size

    # First creation of positions
    if not variables.POINTS_ID:
        for (x, y), (vx, vy) in zip(reynolds.birds, reynolds.velocities):
            if not params.TRIANGLES:
                pid = canvas.create_oval(
                    x - size,
                    y - size,
                    x + size,
                    y + size,
                    fill=fill,
                    outline=outline,
                    tag="bird",
                )
            else:
                angle = math.atan2(vy, vx)
                cos_a, sin_a = math.cos(angle), math.sin(angle)

                tip_x = x + cos_a * triangle_size
                tip_y = y + sin_a * triangle_size

                left_x = x + (cos_a * cos_150 - sin_a * sin_150) * triangle_width
                left_y = y + (sin_a * cos_150 + cos_a * sin_150) * triangle_width

                right_x = x + (cos_a * cos_150 + sin_a * sin_150) * triangle_width
                right_y = y + (sin_a * cos_150 - cos_a * sin_150) * triangle_width

                pid = canvas.create_polygon(
                    tip_x,
                    tip_y,
                    left_x,
                    left_y,
                    right_x,
                    right_y,
                    fill=fill,
                    outline=outline,
                    tag="bird",
                )
            variables.POINTS_ID.append(pid)

    # Update positions
    for pid, (x, y), (vx, vy) in zip(
        variables.POINTS_ID, reynolds.birds, reynolds.velocities
    ):
        if not params.TRIANGLES:
            canvas.coords(pid, x - size, y - size, x + size, y + size)
        else:
            angle = math.atan2(vy, vx)
            cos_a, sin_a = math.cos(angle), math.sin(angle)

            tip_x = x + cos_a * triangle_size
            tip_y = y + sin_a * triangle_size

            left_x = x + (cos_a * cos_150 - sin_a * sin_150) * triangle_width
            left_y = y + (sin_a * cos_150 + cos_a * sin_150) * triangle_width

            right_x = x + (cos_a * cos_150 + sin_a * sin_150) * triangle_width
            right_y = y + (sin_a * cos_150 - cos_a * sin_150) * triangle_width

            canvas.coords(pid, tip_x, tip_y, left_x, left_y, right_x, right_y)

    # Delete uncesserary points
    if len(variables.POINTS_ID) > len(reynolds.birds):
        for pid in variables.POINTS_ID[len(reynolds.birds) :]:
            canvas.delete(pid)
        variables.POINTS_ID = variables.POINTS_ID[: len(reynolds.birds)]
    # Create missing points
    elif len(variables.POINTS_ID) < len(reynolds.birds):
        for x, y in reynolds.birds[len(variables.POINTS_ID) :]:
            if not params.TRIANGLES:
                pid = canvas.create_oval(
                    x - size,
                    y - size,
                    x + size,
                    y + size,
                    fill=fill,
                    outline=outline,
                    tag="bird",
                )
            else:
                angle = math.atan2(0, 1)  # orientation par défaut
                cos_a, sin_a = math.cos(angle), math.sin(angle)

                tip_x = x + cos_a * triangle_size
                tip_y = y + sin_a * triangle_size

                left_x = x + (cos_a * cos_150 - sin_a * sin_150) * triangle_width
                left_y = y + (sin_a * cos_150 + cos_a * sin_150) * triangle_width

                right_x = x + (cos_a * cos_150 + sin_a * sin_150) * triangle_width
                right_y = y + (sin_a * cos_150 - cos_a * sin_150) * triangle_width

                pid = canvas.create_polygon(
                    tip_x,
                    tip_y,
                    left_x,
                    left_y,
                    right_x,
                    right_y,
                    fill=fill,
                    outline=outline,
                    tag="bird",
                )
            variables.POINTS_ID.append(pid)


def maximize_minimize(force):
    if is_maximized():
        root.state("normal")
        try:
            root.attributes("-zoomed", False)
        except tk.TclError:
            pass
        if not force:
            params.WIDTH = const.WIDTH_BEFORE_MAXIMIZED
            params.HEIGHT = const.HEIGHT_BEFORE_MAXIMIZED
    else:
        const.WIDTH_BEFORE_MAXIMIZED = params.WIDTH
        const.HEIGHT_BEFORE_MAXIMIZED = params.HEIGHT

        wm = root.tk.call("tk", "windowingsystem")
        if wm == "aqua":  # macOS
            root.attributes("-fullscreen", True)
        elif wm == "win32":  # Windows
            root.state("zoomed")
        else:  # Linux
            try:
                root.attributes("-zoomed", True)
            except tk.TclError:
                root.attributes("-fullscreen", True)
    root.focus_force()
    root.focus_set()


def show_tip(widget, text, event=None, dx=10, dy=10, wraplength=200):

    if variables.TIP_WINDOW is not None:
        try:
            variables.TIP_WINDOW.destroy()
        except:
            pass
        variables.TIP_WINDOW = None

    if not text:
        return

    if event:
        x = widget.winfo_rootx() + event.x + dx
        y = widget.winfo_rooty() + event.y + dy
    else:
        x = widget.winfo_rootx() + dx
        y = widget.winfo_rooty() + widget.winfo_height() + dy

    variables.TIP_WINDOW = tw = tk.Toplevel(widget)
    tw.wm_overrideredirect(True)
    tw.wm_geometry(f"+{x}+{y}")
    label = tk.Label(
        tw,
        text=text,
        background=variables.INFO_COLOR,
        relief="solid",
        borderwidth=1,
        font=(params.FONT_TYPE, params.FONT_SIZE),
        wraplength=wraplength,
    )
    label.pack(ipadx=4, ipady=2)


def hide_tip(event=None):
    if variables.TIP_WINDOW:
        variables.TIP_WINDOW.destroy()
        variables.TIP_WINDOW = None


def add_canvas_tooltip(item, text):
    canvas.tag_bind(item, "<Enter>", lambda e: show_tip(canvas, text, e))
    canvas.tag_bind(item, "<Leave>", hide_tip)


def add_widget_tooltip(widget, text):
    widget.bind("<Enter>", lambda e: show_tip(widget, text))
    widget.bind("<Leave>", hide_tip)


def on_resize(on_other_key, start_repeat, stop_repeat, event):
    if variables.TRANS_HIDDEN:
        params.WIDTH = max(
            event.width - variables.WIDTH_PARAMS - variables.WIDTH_CONTROLS,
            params.WIDTH_MIN,
        )
        params.HEIGHT = max(event.height, variables.HEIGHT_PARAMS_CONTROLS)
        move_birds(False, False)
        draw_birds()
        draw_box()
        draw_status_overlays()
        variables.TRANS_HIDDEN = False
        return
    # TODO BUGIFX
    params.WIDTH = max(
        event.width - variables.WIDTH_PARAMS - variables.WIDTH_CONTROLS - 2,
        params.WIDTH_MIN,
    )
    params.HEIGHT = max(event.height, variables.HEIGHT_PARAMS_CONTROLS)
    # params.WIDTH = max(event.width - variables.WIDTH_PARAMS - variables.WIDTH_CONTROLS,params.WIDTH_MIN)
    # params.HEIGHT = max(event.height,const.HEIGHT_PARAMS_CONTROLS_DEFAULT)
    move_birds(False, False)
    draw_panels(True, on_other_key, start_repeat, stop_repeat)
    draw_birds()
    draw_box()
    draw_status_overlays()



from importlib.metadata import version

VERSION_PROG = version("pybirdsreynolds")
SHORT_DESCRIPTION = (
    "pybirdsreynolds is an interactive simulation of the Reynolds Boids model."
)
MARGIN = 1
WIDTH_CONTROLS_DEFAULT = 70
WIDTH_PARAMS_DEFAULT = 350
HEIGHT_PARAMS_CONTROLS_DEFAULT = 600
FONT_TYPE_LIST = [
    "Nota Sans Mono",
    "Consolas",
    "Courier",
    "Ubuntu Sans Mono",
    "DejaVu Sans Mono",
    "Courier New",
    "Liberation Mono",
    "Monaco", 
    "Menlo",
    "PT Mono",
    "Fira Mono", 
    "JetBrains Mono",
    "Source Code Pro", 
    "Inconsolata",
]



PAUSED = True
BLINK_STATE = True
FPS = False
HIDDEN = False
TRANS_HIDDEN = False
SHIFT_PRESSED = False
REPEATING = {"active": False, "job": None}
CANVAS_BG = None
FILL_COLOR = None
OUTLINE_COLOR = None
INFO_COLOR = None
WIDTH_CONTROLS = 70
WIDTH_PARAMS = 350
HEIGHT_PARAMS_CONTROLS = 600
FRAME_COUNT = 0
FPS_VALUE = 0
LAST_TIME = 0
COUNT = 0
POINTS_ID = []
WIDTH_BEFORE_MAXIMIZED = 0
HEIGHT_BEFORE_MAXIMIZED = 0
TIP_WINDOW = None
AVG_X = 0
AVG_Y = 0
AVG_SPEED_ABS = 0
AVG_VX = 0
AVG_VY = 0
AVERAGE = False
HUNTER = False
